/******************************************************************************
*
* project name:    SOKOMIZE
* initial date:    10/05/2000
* author:          thomas.nussbaumer@gmx.net
* description:     Randomized Sokoban Clone
*
* -----------------------------------------------------------------------------
* Revision History
* -----------------------------------------------------------------------------
*
* v1.10: 18/11/2002
*        changes for V200/AMS 2.08/TIGCC 0.94
*
* v1.00: 06/06/2000
*        (1) call of idle() function in GetInputKeyWithAPD() removed, because
*            it interferes with grayscale routines
*        (2) SHIFT+direction key scrolling added and F1 scroll toggle removed
*        (3) QUICK SAVE/RESTORE feature added ([STO] and [CLEAR])
*        (4) turn calculator off using [DIAMOND][OFF] or [SHIFT][OFF] added
*        (5) take back last move ([BACKSPACE]) added
*        (6) end of level handling modified
*
* v0.9a: 30/05/2000
*        menu added and commandline parameters handling removed
*        minor problems in file parsing fixed
*        APD timer gets now restarted after key press
*
* v0.99: 29/05/2000
*        changes due to TIGCC standard library 2.0:
*        (1) usage of SAVE_SCREEN & USE_GRAY_PATCH precompiler defines
*        (2) usage of internal random number generator
*        (3) usage of key definitions from compat.h
*        (4) title screen added
*        (5) own implementation of GetInputKeyWithAPD() replaces ngetchx()
*        (6) redraws dirty squares before displaying level-finished message
*        (7) make sure player is visible in preset level mode
*
* v0.97: 25/05/2000
*        usage of predefined levels by using streaming input from text
*        variable (feature of TIGCC standard library 2.0)
*
* v0.96: 17/05/2000
*        scrolls screen now in such a way that player can see to where he moves
*        boxes (player character is two squares away from the border)
*
* v0.95: 15/05/2000
*        main parameters can now be specified on the commandline
*
* v0.92: 12/05/2000
*        heavily documented so it may be used as "tutorial" sample
*
* v0.91: 11/05/2000
*        completeness check fixed
*
* v0.90: 10/05/2000
*        initial version
*
*******************************************************************************/

// KEYS:
//
// [UP]            ... move up
// [DOWN]          ... move down
// [LEFT]          ... move left
// [RIGHT]         ... move right
// [SHIFT]+[UP]    ... scroll up
// [SHIFT]+[DOWN]  ... scroll down
// [SHIFT]+[LEFT]  ... scroll left
// [SHIFT]+[RIGHT] ... scroll right
// [ESCAPE]        ... quit game
// [STO]           ... during game: quick save
// [CLEAR]         ... during game: quick restore
// [BACKSPACE]     ... take back last move
// [SHIFT]  +[ON]  ... turn off calculator during level
// [DIAMOND]+[ON]  ... turn off calculator during level
//
//=============================================================================

#define SAVE_SCREEN
#define OPTIMIZE_ROM_CALLS
#define NO_EXIT_SUPPORT

#if defined(USE_TI89)
#define C89_92(a,b)  (a)
#else
#define C89_92(a,b)  (b)
#endif

#include <tigcclib.h>

#define SOKOMIZE_VERSION "v1.20"


//#include "title.h"      // data of title screen including decompression routine
#include "menu.c"       // data of main menu and handling routines
#include "logo.c"

//-----------------------------------------------------------------------------
// MAJOR CONTROL DEFINES FOR LEVELS (DEFAULTS)
//-----------------------------------------------------------------------------
#define NR_COLUMNS   22   // number of columns
#define NR_LINES     22   // number of lines
#define RANDOM_BOXES 50   // number of tries to distribute a random box
#define RANDOM_DESTS 10   // number of randomly distributed destination squares
#define RANDOM_SOLID 10   // percentage of solid blocks within level


#define KEY_ESCAPE       264
#define KEY_ENTER         13
#define KEY_BACK         257    // [BACKSPACE]
#define KEY_SAVE         258    // [STO]
#define KEY_OFF_1        4363
#define KEY_RESTORE      263    // [CLEAR]
#define VISIBLE_COLUMNS  C89_92(10,15)
#define VISIBLE_LINES    C89_92(6,8)
#define SHIFT_MASK       C89_92(0x20,0x04)  // row mask for SHIFT key
#define KEY_OFF_2        C89_92(16651,8459)



//-----------------------------------------------------------------------------
// background tiles data and defines
//
// (background graphics consists of 2 planes a 16 WORDs)
//-----------------------------------------------------------------------------
#define BG_NONE     0xff
#define BG_ROCK     0
#define BG_FLOOR    1
#define BG_DEST     2

unsigned short img_bg[6*16] = {
// SOLID ROCK
0xFC7F,0x94DD,0xBDF9,0x1A76,0xF78C,0xE77D,0xA6DB,0x1D7B,0x6D37,0xBCC0,0xF3FF,0x0D4E,0x3ECC,0xBBD9,0x3739,0x086A,
0x0380,0x6302,0x6306,0xE78F,0x087F,0x1882,0x5984,0xE384,0xC2CC,0xC3BF,0x0E00,0xFE31,0xC033,0x4436,0xCCF6,0xFFF5,
// NORMAL FLOOR
0x0380,0x6302,0x6306,0xE78F,0x087F,0x1882,0x5984,0xE384,0xC2CC,0xC3BF,0x0E00,0xFE31,0xC033,0x4436,0xCCF6,0xFFF5,
0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
// DESTINATION SQUARE
0xffff,0xe303,0xe307,0xe78f,0x887f,0x9883,0xd985,0xe385,0xc2cc,0xc3bf,0x8e01,0xfe31,0xc033,0xc437,0xccf7,0xffff,
0xffff,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0xffff
};


//-----------------------------------------------------------------------------
// item graphics data and defines
//
// (item graphics consists of 2 planes + 1 mask a 16 WORDs)
//-----------------------------------------------------------------------------
#define ITEM_NONE   0xff
#define ITEM_PLAYER 0
#define ITEM_BOX    1

unsigned short img_item[6*16] = {
//PLAYER (plane0/plane1/mask)
0x0180,0x0240,0x0EF0,0x1088,0x1418,0x2624,0x2C6C,0x5C7A,0x4E72,0x599A,0x3AD4,0x1DF0,0x16D0,0x2458,0x7C64,0x0000,
0x0180,0x0240,0x0E70,0x1248,0x1188,0x2464,0x2424,0x4C32,0x4FF2,0x5C3A,0x3994,0x13D0,0x1290,0x3C68,0x4444,0x0000,
0xFE7F,0xFC3F,0xF00F,0xE007,0xE007,0xC003,0xC003,0x8001,0x8001,0x8001,0xC00B,0xE00F,0xE10F,0xC387,0x8383,0xFFFF,
//BOX (plane0/plane1/mask)
0x0000,0x7FFE,0x6A82,0x5542,0x400E,0x6156,0x6002,0x4082,0x4052,0x400E,0x60AE,0x6A02,0x5552,0x400E,0x7FFE,0x0000,
0x0000,0x7FFE,0x557E,0x4ABE,0x6002,0x71FE,0x4EFE,0x4E7E,0x51AE,0x7F02,0x7FFE,0x55FE,0x4AAE,0x6002,0x7FFE,0x0000,
0xFFFF,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0x8001,0xFFFF
};


//-----------------------------------------------------------------------------
// structure describing one square of the map
//-----------------------------------------------------------------------------
typedef struct {
    unsigned char background;  // ID of background
    unsigned char item;        // ID of item on this square or ITEM_NONE
    unsigned char dirty;       // 1 if square should be redrawn (otherwise 0)
} sq_type;


#define RANDOM_MODE 0


//-----------------------------------------------------------------------------
// game globals
//-----------------------------------------------------------------------------
sq_type**       map        = 0;  // the actual map ([NR_LINES][NR_COLUMNS])
unsigned short**  offsets    = 0;  // offsets of the visible squares ([VISIBLE_COLUMNS][VISIBLE_LINES])
unsigned short*   plane1     = 0;  // startadress of plane 1
unsigned short*   plane2     = 0;  // startadress of plane 2

unsigned short    startx     = 0;  // x value within the map of upper left corner square
unsigned short    starty     = 0;  // y value within the map of upper left corner square
unsigned short    startx_old = 0;  // used to store old startx if scrolling mode is activated
unsigned short    starty_old = 0;  // used to store old starty if scrolling mode is activated
unsigned short    playerx    = 1;  // the player's x value within the map
unsigned short    playery    = 1;  // the player's y value within the map
short             scrollmode = 0;  // 1 if scrollmode is on otherwise 0
short             boxes_rand = RANDOM_BOXES; // number of random boxes
short             dest_rand  = RANDOM_DESTS; // number of random destination squares
short             solid_rand = RANDOM_SOLID; // percentage of solid blocks within level
short             level_nr   = 0;            // LEVEL number
short             file_nr    = 0;            // FILE number to use
short             play_mode = RANDOM_MODE;   // mode of play


//-----------------------------------------------------------------------------
// backup of global variables (used for smart take back feature)
//-----------------------------------------------------------------------------
sq_type**       back_map        = 0;
unsigned short    back_startx     = 0;
unsigned short    back_starty     = 0;
unsigned short    back_startx_old = 0;
unsigned short    back_starty_old = 0;
unsigned short    back_playerx    = 1;
unsigned short    back_playery    = 1;
short             back_scrollmode = 0;

//-----------------------------------------------------------------------------
// backup of global variables (used for back one move)
//-----------------------------------------------------------------------------
sq_type**       back1_map        = 0;
unsigned short    back1_startx     = 0;
unsigned short    back1_starty     = 0;
unsigned short    back1_startx_old = 0;
unsigned short    back1_starty_old = 0;
unsigned short    back1_playerx    = 1;
unsigned short    back1_playery    = 1;
short             back1_scrollmode = 0;



/*===========================================================================*/
/* get input key                                                             */
/*                                                                           */
/* replacement for ngetchx() and/or GKeyIn(), because:                       */
/*                                                                           */
/* (1) ngetchx() won't check the APD timer value                             */
/* (2) GKeyIn() destroys the display due to setting the status display       */
/*                                                                           */
/* additionally special behaviour:                                           */
/*                                                                           */
/* if SHIFT key is pressed alone, function returns KEY_SHIFT                 */
/* if SHIFT key is released, function returns 0                              */
/*                                                                           */
/* this is necessary to get the SHIFT+direction key scrolling working        */
/*                                                                           */
/*===========================================================================*/
short shift_toggle = 0;

short GetInputKeyWithAPD() {
    void *kbq = kbd_queue();
    short key;
    short gray = GrayCheckRunning();

    OSTimerRestart(APD_TIMER);
    //---------------------------------------
    // wait until we get a key
    // if the APD timer expires -> turn
    // calculator off
    //---------------------------------------
    while (OSdequeue(&key, kbq)) {
        if (!gray) idle();
        else pokeIO(0x600005,0b10111);
        if (OSTimerExpired(APD_TIMER)) off();
        else {
            if (_rowread(0) & SHIFT_MASK) {
                key = KEY_SHIFT;
                shift_toggle = 1;
                break;
            }
            else if (shift_toggle) {
                key = 0;
                break;
            }
        }
    }

    OSTimerRestart(APD_TIMER);

    shift_toggle = key & KEY_SHIFT;

    return key & ~(unsigned short)0x800; // map out key repeat flag
}



/*===========================================================================*/
/* initializes memory                                                        */
/*                                                                           */
/* --- Description of the memory layout of 2-dimensional arrays in C ---     */
/*                                                                           */
/* Suppose we have the following static integer array: int MyArray[3][5]     */
/* and want change it into a dynamically allocated array.                    */
/*                                                                           */
/* (1): replace declaration "int MyArray[3][5];" with "int ** MyArray;"      */
/*                                                                           */
/* (2): allocate first dimension which is a array of length 3 which holds    */
/*      pointers to int arrays:                                              */
/*                                                                           */
/*      MyArray = (int**)malloc(sizeof(int*)*3);                             */
/*      if (MyArray == 0) {                                                  */
/*         // handle "OutOfMemory" here ...                                  */
/*      }                                                                    */
/*                                                                           */
/* (3): allocate arrays for the second dimension:                            */
/*                                                                           */
/*      (a) simple way using a loop:                                         */
/*                                                                           */
/*      for (i=0;i<3;i++) {                                                  */
/*         MyArray[i] = (int*)malloc(sizeof(int)*5);                         */
/*         if (MyArray[0] == 0) {                                            */
/*            // handle "OutOfMemory" here ...                               */
/*         }                                                                 */
/*      }                                                                    */
/*                                                                           */
/*      (b) a much better way is to just allocate 1 array that is            */
/*      large enough to hold all 3 arrays and split it up afterwards         */
/*      in three pieces like this:                                           */
/*                                                                           */
/*      MyArray[0] = (int*)malloc(3*sizeof(int)*5);                          */
/*      if (MyArray[0] == 0) {                                               */
/*         // handle "OutOfMemory" here ...                                  */
/*      }                                                                    */
/*                                                                           */
/*      for (i=1;i<3;i++) {                                                  */
/*         MyArray[i] = MyArray[0] + i * 5;                                  */
/*      }                                                                    */
/*                                                                           */
/*      This way is better, because the allocated space of the second        */
/*      dimension can be free'd later in one piece, too.                     */
/*      Additionally each allocated block will use a system handle, so       */
/*      keeping the number of allocated blocks small is allways a good       */
/*      idea.                                                                */
/*                                                                           */
/*===========================================================================*/
short InitMemory(void) {
    short i;

    //--------------------------------------------------
    // allocate the arrays of pointers
    //--------------------------------------------------
    if (!(map       = (sq_type**)malloc(sizeof(sq_type*)*NR_LINES))) return 0;
    if (!(back_map  = (sq_type**)malloc(sizeof(sq_type*)*NR_LINES))) return 0;
    if (!(back1_map = (sq_type**)malloc(sizeof(sq_type*)*NR_LINES))) return 0;
    if (!(offsets   = (unsigned short**)malloc(sizeof(unsigned short*)*VISIBLE_LINES))) return 0;

    //---------------------------------------------------
    // now allocate space for 2nd dimensions in one piece
    // and split it afterwards into the single pointers
    //--------------------------------------------------
    if (!(map[0]       = (sq_type*)malloc(sizeof(sq_type)*NR_COLUMNS*NR_LINES))) return 0;
    if (!(back_map[0]  = (sq_type*)malloc(sizeof(sq_type)*NR_COLUMNS*NR_LINES))) return 0;
    if (!(back1_map[0] = (sq_type*)malloc(sizeof(sq_type)*NR_COLUMNS*NR_LINES))) return 0;
    if (!(offsets[0]   = (unsigned short*)malloc(sizeof(unsigned short)*VISIBLE_COLUMNS*VISIBLE_LINES))) return 0;

    for (i=1;i<NR_LINES;i++) {
        map[i]       = map[0]       + i*NR_COLUMNS;
        back_map[i]  = back_map[0]  + i*NR_COLUMNS;
        back1_map[i] = back1_map[0] + i*NR_COLUMNS;
    }

    for (i=1;i<VISIBLE_LINES;i++) {
        offsets[i] = offsets[0] + i*VISIBLE_COLUMNS;
    }

    return 1;
}



/*===========================================================================*/
/* frees memory                                                              */
/*===========================================================================*/
void FreeMemory(void) {
    if (map) {
        if (map[0]) free(map[0]);
        free(map);
        map = 0;
    }
    if (back_map) {
        if (back_map[0]) free(back_map[0]);
        free(back_map);
        back_map = 0;
    }
    if (back1_map) {
        if (back1_map[0]) free(back1_map[0]);
        free(back1_map);
        back1_map = 0;
    }
    if (offsets) {
        if (offsets[0]) free(offsets[0]);
        free(offsets);
        offsets = 0;
    }
}



/*===========================================================================*/
/* position file pointer on a given level number                             */
/* returns 1 if successfully positionated                                    */
/*===========================================================================*/
short FindLevel(FILE* fp, short nr) {
    char buf[256];
    short  actual_nr = 0;
    rewind(fp);

    if (nr == 0) return 1;

    do {
        if (!fgets(buf,256,fp)) return 0;
        if (buf[0]==';') {
            actual_nr++;
            if (actual_nr == nr) return 1;
        }
    }
    while (1);
}



/*===========================================================================*/
/* loads a level from a file                                                 */
/* returns 1 if level successfully loaded                                    */
/*===========================================================================*/
short LoadLevel(FILE* fp) {
    char   buf[256];
    fpos_t startpos;
    short  line_count   = 0; // number of lines of the level
    short  column_count = 0; // number of columns in the level
    short  length;
    short  line;
    short  column;

    //------------------------------------
    // store original level start position
    //------------------------------------
    fgetpos(fp, &startpos);

    //---------------------------------
    // try to evaluate level extensions
    //---------------------------------
    buf[0] = 0;
    do {
         if (!fgets(buf, 256, fp)) break;

         if (buf[0] == ';') break;

         length = strlen(buf);
         line_count++;
         if (length > column_count) column_count = length;
    }
    while (1);

    //---------------------------------------------------------------
    // check if the level is valid and correct extensions if possible
    //---------------------------------------------------------------
    if (!line_count || !column_count) return 0;

    if (line_count   > NR_LINES-2)   line_count   = NR_LINES - 2;
    if (column_count > NR_COLUMNS-2) column_count = NR_COLUMNS - 2;

    //------------------------------------------------------------------------
    // loop across complete level and set the type (background) of the squares
    //------------------------------------------------------------------------
    for (line=0;line<NR_LINES;line++) {
        for (column=0;column<NR_COLUMNS;column++) {
            map[line][column].item = ITEM_NONE;

            //------------------------------------------------------
            // border squares of the level will be set to solid rock
            //------------------------------------------------------
            if (line == 0 || line >= line_count ||
                column == 0 || column >= column_count)
            {
                map[line][column].background = BG_ROCK;
            }
            else {
                map[line][column].background = BG_FLOOR;
            }
        }
    }


    //-------------------------------
    // set filepointer to level start
    //-------------------------------
    fsetpos(fp,&startpos);

    for (line=0;line<line_count;line++) {
        fgets(buf,256,fp);
        length = strlen(buf);
        for (column=0;column<length;column++) {
            switch(buf[column]) {
                case ' ': map[line+1][column+1].background = BG_FLOOR;
                          break;
                case 'W': map[line+1][column+1].background = BG_ROCK;
                          break;
                case 'B': map[line+1][column+1].background = BG_FLOOR;
                          map[line+1][column+1].item       = ITEM_BOX;
                          break;
                case 'b': map[line+1][column+1].item       = ITEM_BOX;
                          map[line+1][column+1].background = BG_DEST;
                          break;
                case 'D': map[line+1][column+1].background = BG_DEST;
                          break;
                case 'P': map[line+1][column+1].background = BG_FLOOR;
                          map[line+1][column+1].item       = ITEM_PLAYER;
                          playery = line+1;
                          playerx = column+1;
                          break;
                case 'p': map[line+1][column+1].item       = ITEM_PLAYER;
                          map[line+1][column+1].background = BG_DEST;
                          playery = line+1;
                          playerx = column+1;
                          break;
            }
        }
    }

    return 1;
}



/*===========================================================================*/
/* blits a 16x16 tile to both planes                                         */
/*                                                                           */
/* line/column = logical square coordinates on screen                        */
/* src         = source of image (30 bytes for plane1 + 30 bytes for plane2) */
/*===========================================================================*/
void Blit16x16Full(unsigned short line,unsigned short column, unsigned short* src) {
    unsigned short offset = offsets[line][column];
    unsigned short *dest1 = plane1 + offset;
    unsigned short *dest2 = plane2 + offset;
    unsigned short *src2  = src + 16;

    short i;

    for (i=0;i<16;i++) {
        *dest1 = *src++;
        *dest2 = *src2++;
        dest1+=15;
        dest2+=15;
    }
}



/*===========================================================================*/
/* blits a 16x16 tile to both planes using a mask                            */
/*                                                                           */
/* line/column = logical square coordinates on screen                        */
/* src         = source of image (30 bytes for plane1 +                      */
/*                                30 bytes for plane2 +                      */
/*                                30 bytes for mask)                         */
/*===========================================================================*/
void Blit16x16Mask(short line,short column, unsigned short* src) {
    unsigned short offset = offsets[line][column];
    unsigned short *dest1 = plane1 + offset;
    unsigned short *dest2 = plane2 + offset;
    unsigned short *src2  = src + 16;
    unsigned short *mask  = src + 32;
    short i;

    for (i=0;i<16;i++) {
        *dest1 &= *mask;
        *dest1 |= *src++;
        *dest2 &= *mask++;
        *dest2 |= *src2++;
        dest1+=15;
        dest2+=15;
    }
}



/*===========================================================================*/
/* initializes level map                                                     */
/*===========================================================================*/
void InitMap(void) {
    short line,column,i;

    //------------------------------------------------------------------------
    // loop across complete level and set the type (background) of the squares
    //------------------------------------------------------------------------
    for (line=0;line<NR_LINES;line++) {
        for (column=0;column<NR_COLUMNS;column++) {
            map[line][column].item = ITEM_NONE;

            //------------------------------------------------------
            // border squares of the level will be set to solid rock
            //------------------------------------------------------
            if (line == 0 || line == NR_LINES-1 ||
                column == 0 || column == NR_COLUMNS-1)
            {
                map[line][column].background = BG_ROCK;
            }
            else {
                short i = random(100);

                if (i < solid_rand) {
                    //----------------------------------
                    // squares will be set to solid rock
                    //----------------------------------
                    map[line][column].background = BG_ROCK;
                }
                else {
                    //------------------------------------------------
                    // rest of the squares will be set to normal floor
                    //------------------------------------------------
                    map[line][column].background = BG_FLOOR;
                }
            }
        }
    }


    //------------------------------------------------------------------------
    // distribute destination fields, but make sure that no destination square
    // will be on a border square
    //------------------------------------------------------------------------
    for (i=0;i<dest_rand;i++) {
        line   = (random(NR_LINES - 2))   + 1;
        column = (random(NR_COLUMNS - 2)) + 1;
        map[line][column].background = BG_DEST;
    }


    //------------------------------------------------------------------------
    // distribute boxes, but make sure that no box is set on a border square
    // if a box will be set on a solid rock square -> change square type to
    // normal floor
    //------------------------------------------------------------------------
    for (i=0;i<boxes_rand;i++) {
        line   = (random(NR_LINES - 2))   + 1;
        column = (random(NR_COLUMNS - 2)) + 1;
        if (map[line][column].background == BG_ROCK) {
            map[line][column].background = BG_FLOOR;
        }

        map[line][column].item = ITEM_BOX;
    }

}



/*===========================================================================*/
/* initializes offsets array                                                 */
/*                                                                           */
/* the offsets array is used as lookup table for the starting addresses of   */
/* the visible squares within the planes                                     */
/* the calculations could be done on-the-fly without the offsets table, too, */
/* but is precalculated to speed up the drawing a little bit                 */
/*===========================================================================*/
void InitOffsets(void) {
    unsigned short line,column;

    for (line = 0; line < VISIBLE_LINES; line ++) {
        for (column = 0; column < VISIBLE_COLUMNS; column++) {
            offsets[line][column] = line * 15 * 16 + column;
        }
    }
}



/*===========================================================================*/
/* draw a square                                                             */
/*                                                                           */
/* drawing a single square is done in two steps to keep things simple:       */
/* (1) the background is drawn                                               */
/* (2) if there is an item on this square, it will be drawn, too             */
/*                                                                           */
/* this routine may be improved, because if there is an item on this square  */
/* its not necessary to draw the complete background                         */
/*===========================================================================*/
void DrawSquare(short line, short column) {
    sq_type* act_sq = &(map[starty+line][startx+column]);

    Blit16x16Full(line,column,&img_bg[act_sq->background*32]);

    if (act_sq->item != ITEM_NONE) {
        Blit16x16Mask(line,column,&img_item[act_sq->item*48]);
    }

    act_sq->dirty = 0;
}



/*===========================================================================*/
/* draw scroll mode indicator                                                */
/*                                                                           */
/* the scroll mode indicator is a tiny black square in the top left corner   */
/* of the screen, just to help the user remember which mode is active        */
/*===========================================================================*/
void OutputMode() {
    short i;
    unsigned char* dest1 = (unsigned char*)plane1;
    unsigned char* dest2 = (unsigned char*)plane2;
    if (scrollmode) {
        //-----------------------------------------------------
        // draw indicator which is 4 black pixels wide and high
        //-----------------------------------------------------
        for (i=0;i<4;i++,dest1+=30,dest2+=30) {
            *dest1 |= 0xf0;
            *dest2 |= 0xf0;
        }
    }
}



/*===========================================================================*/
/* special effect clear screen                                               */
/*                                                                           */
/* it should not be too complicated to figure out how this routine works ;-) */
/*===========================================================================*/
void FXClearScreen() {
    short height  = ScrRect->xy.y1;
    short height2 = height/2;
    short width   = ScrRect->xy.x1;
    short width2  = width/2;
    short i,j;

    //---------------------
    // clear in y direction
    //---------------------
    for (i=0;i<=height2;i++) {
        DrawLine(0,i,width,i,A_REPLACE);
        DrawLine(0,height-i,width,height-i,A_REPLACE);
        DrawLine(0,i-1,width,i-1,A_REVERSE);
        DrawLine(0,height-i+1,width,height-i+1,A_REVERSE);
        for (j=0;j<=10000;j++);
    }

    //-----------------------------
    // clear in x direction
    //-----------------------------
    for (i=0;i<=width2;i++) {
        DrawPix(i,height2,A_REVERSE);
        DrawPix(width-i,height2,A_REVERSE);
        for (j=0;j<=3000;j++);
    }
}



/*===========================================================================*/
/* helper function for detecting a YES key press                             */
/*===========================================================================*/
short isYesKey(short key) {
    if (key == 'Y'       ||
        key == 'y'       ||
        key == KEY_ENTER)
    {
        return TRUE;
    }
    else {
        return FALSE;
    }
}



/*===========================================================================*/
/* helper function for detecting a NO key press                              */
/*===========================================================================*/
short isNoKey(short key) {
    if (key == 'N'        ||
        key == 'n'        ||
        key == KEY_ESCAPE
#if defined(USE_TI89)
        || key == '6'
#endif
       )
    {
        return TRUE;
    }
    else {
        return FALSE;
    }
}



/*===========================================================================*/
/* handles a request dialog (returns 1 if YES is pressed)                    */
/*===========================================================================*/
short RequestDialog(const char* request) {
    short read;
    short yes;
    short no;
    short i;
    short j;
    short sy;
    short p;
    short oldfont = FontGetSys();

    LCD_BUFFER LCD_tmp;
    LCD_BUFFER LCD_tmp2;

    SetPlane(0);
    BitmapGet(ScrRect,LCD_tmp);
    SetPlane(1);
    BitmapGet(ScrRect,LCD_tmp2);

    sy = ScrRect->xy.y1/2 - 2;

    for (i=0;i<=10;i++) {
        for (p=0;p<2;p++) {
           SetPlane(p);
           DrawClipRect(MakeWinRect(0,sy-i,ScrRect->xy.x1-1,sy+i),ScrRect,1);
           DrawClipRect(MakeWinRect(1,sy-i+1,ScrRect->xy.x1-2,sy+i-1),ScrRect,0);
        }
        for (j=0;j<=4000;j++);
    }

    FontSetSys(F_6x8);

    for (p=0;p<2;p++) {
       SetPlane(p);
       DrawStr((ScrRect->xy.x1-DrawStrWidth(request,F_6x8))/2,sy-3,request,A_REPLACE);
    }

    do {
        read = GetInputKeyWithAPD();
        yes  = isYesKey(read);
        no   = isNoKey(read);
    }
    while (!yes && !no);

    SetPlane(0);
    BitmapPut(0,0,LCD_tmp,ScrRect,A_REPLACE);
    SetPlane(1);
    BitmapPut(0,0,LCD_tmp2,ScrRect,A_REPLACE);

    FontSetSys(oldfont);
    return yes;
}



/*===========================================================================*/
/* checks if this level is complete                                          */
/*                                                                           */
/* loops across complete level and returns 0 if it finds a destination       */
/* square with no box on it                                                  */
/*===========================================================================*/
short LevelCompleted() {
    short line,column;

    for (line=0;line<NR_LINES;line++) {
        for (column=0;column<NR_COLUMNS;column++) {
           if (map[line][column].background == BG_DEST) {
               if (map[line][column].item != ITEM_BOX) return 0;
           }
        }
    }
    return 1;
}



/*===========================================================================*/
/* redraws squares which should be updated                                   */
/*===========================================================================*/
void RedrawDirtySquares() {
    short line,column;
    for (line = 0; line < VISIBLE_LINES; line++) {
        for (column = 0; column < VISIBLE_COLUMNS; column++) {
            if (map[starty+line][startx+column].dirty) {
                DrawSquare(line,column);
                if (line == 0 && column == 0) OutputMode();
            }
        }
    }
}



/*===========================================================================*/
/* backup level (one move back feature)                                      */
/*===========================================================================*/
void BackupLevel() {
    memcpy(back_map[0],map[0],sizeof(sq_type)*NR_COLUMNS*NR_LINES);
    back_startx     = startx;
    back_starty     = starty;
    back_startx_old = startx_old;
    back_starty_old = starty_old;
    back_playerx    = playerx;
    back_playery    = playery;
    back_scrollmode = scrollmode;
}



/*===========================================================================*/
/* backup level (quick restore feature)                                      */
/*===========================================================================*/
void Backup1Level() {
    memcpy(back1_map[0],map[0],sizeof(sq_type)*NR_COLUMNS*NR_LINES);
    back1_startx     = startx;
    back1_starty     = starty;
    back1_startx_old = startx_old;
    back1_starty_old = starty_old;
    back1_playerx    = playerx;
    back1_playery    = playery;
    back1_scrollmode = scrollmode;
}



/*===========================================================================*/
/* generate and play a level                                                 */
/* returns 0 ... if user wants to abort                                      */
/* returns 1 ... if user wants to play the next level                        */
/* returns 2 ... if user wants to restart preset level                       */
/*===========================================================================*/
short GenerateAndPlay() {
    short   line,column;
    short   input;
    short   finish         = 0;
    short   screen_refresh = 1;
    FILE* fp             = 0;
    short   backup_done    = 0;

    InitOffsets();
    startx     = 0;
    starty     = 0;
    startx_old = 0;
    starty_old = 0;
    scrollmode = 0;

    if (play_mode != RANDOM_MODE) {
        char buf[20];

        sprintf(buf,"m%02d",file_nr);

        if (!(fp = fopen(buf,"r"))) {
            sprintf(buf,"Can't open m%02d",file_nr);
            RequestDialog(buf);
            play_mode = RANDOM_MODE;
        }
        else if (!FindLevel(fp,level_nr)) {
            RequestDialog("No more Levels");
            level_nr = 0;
            return RequestDialog("Start at Level 0?");
        }
        else if (!LoadLevel(fp)) {
            RequestDialog("Invalid Level");
            return RequestDialog("Try next one?");
        }
        else {
            // correct startx and starty if necessary
            if (playery >= starty + VISIBLE_LINES - 3) {
                starty = playery - VISIBLE_LINES + 3;
                if (starty >= NR_LINES - VISIBLE_LINES) {
                    starty = NR_LINES - VISIBLE_LINES;
                }
                starty_old = starty;
            }
            if (playerx >= startx + VISIBLE_COLUMNS - 3) {
                startx = playerx - VISIBLE_COLUMNS + 3;
                if (startx >= NR_COLUMNS - VISIBLE_COLUMNS) {
                    startx = NR_COLUMNS - VISIBLE_COLUMNS;
                }
                startx_old = startx;
            }
        }
        if (fp) fclose(fp);
    }


    if (play_mode == RANDOM_MODE) {
        InitMap();
        playerx    = 1;
        playery    = 1;
        map[playery][playerx].background = BG_FLOOR;    // make sure we start on a floor
        map[playery][playerx].item       = ITEM_PLAYER; // set player in map
    }

    BackupLevel();
    Backup1Level();

    do {
        if (screen_refresh == 1) {
            //------------------
            // draws all squares
            //------------------
            for (line = 0; line < VISIBLE_LINES; line++) {
                for (column = 0; column < VISIBLE_COLUMNS; column++) {
                    DrawSquare(line,column);
                    if (line == 0 && column == 0) OutputMode();
                }
            }
            screen_refresh = 0;
        }
        else {
            //---------------------------------------------
            // draws just squares which are marked as dirty
            //---------------------------------------------
            RedrawDirtySquares();
        }

        input = GetInputKeyWithAPD();

        //------------------------------
        // handle turn off of calculator
        //------------------------------
        if (input == KEY_OFF_1 || input == KEY_OFF_2) {
            off();
            continue;
        }

        //---------------------------------
        // handle SHIFT key quick scrolling
        //---------------------------------
        if (input & KEY_SHIFT) {
            input -= KEY_SHIFT;
            if (!scrollmode) {
                scrollmode  = 1;
                startx_old = startx;
                starty_old = starty;
                map[starty][startx].dirty = 1;
            }
        }
        else {
            if (scrollmode) {
                scrollmode  = 0;
                startx = startx_old;
                starty = starty_old;
                screen_refresh = 1;
            }
        }

        backup_done = 0;

        if (input == KEY_UP) {
            if (scrollmode) {
                if (starty > 0) {
                    starty--;
                    screen_refresh = 1;
                }
                continue;
            }
            if (playery > 0 && map[playery-1][playerx].background != BG_ROCK) {
                if (map[playery-1][playerx].item == ITEM_BOX) {
                    if (playery > 1 &&
                        map[playery-2][playerx].background != BG_ROCK &&
                        map[playery-2][playerx].item != ITEM_BOX)
                    {
                        Backup1Level();
                        backup_done = 1;
                        map[playery-1][playerx].item  = ITEM_NONE;
                        map[playery-2][playerx].item  = ITEM_BOX;
                        map[playery-2][playerx].dirty = 1;
                    }
                    else {
                        continue;
                    }
                }
                if (!backup_done) Backup1Level();

                map[playery][playerx].dirty = 1;
                map[playery][playerx].item  = ITEM_NONE;
                playery--;
                if (playery-starty <= 1) {
                    if (starty > 0) {
                        starty--;
                        screen_refresh = 1;
                    }
                }
                map[playery][playerx].dirty = 1;
                map[playery][playerx].item  = ITEM_PLAYER;
            }
        }

        else if (input == KEY_DOWN) {
            if (scrollmode) {
                if (starty < NR_LINES - VISIBLE_LINES) {
                    starty++;
                    screen_refresh = 1;
                }
                continue;
            }
            if (playery < NR_LINES-1  && map[playery+1][playerx].background != BG_ROCK) {
                if (map[playery+1][playerx].item == ITEM_BOX) {
                    if (playery < NR_LINES-2 &&
                        map[playery+2][playerx].background != BG_ROCK &&
                        map[playery+2][playerx].item != ITEM_BOX)
                    {
                        Backup1Level();
                        backup_done = 1;
                        map[playery+1][playerx].item  = ITEM_NONE;
                        map[playery+2][playerx].item  = ITEM_BOX;
                        map[playery+2][playerx].dirty = 1;
                    }
                    else {
                        continue;
                    }
                }
                if (!backup_done) Backup1Level();
                map[playery][playerx].dirty = 1;
                map[playery][playerx].item  = ITEM_NONE;
                playery++;
                if (playery-starty >= VISIBLE_LINES-2) {
                    if (starty < NR_LINES-VISIBLE_LINES) {
                        starty++;
                        screen_refresh = 1;
                    }
                }
                map[playery][playerx].dirty = 1;
                map[playery][playerx].item  = ITEM_PLAYER;
            }
        }

        else if (input == KEY_LEFT) {
            if (scrollmode) {
                if (startx > 0) {
                    startx--;
                    screen_refresh = 1;
                }
                continue;
            }
            if (playerx > 0 && map[playery][playerx-1].background != BG_ROCK) {
                if (map[playery][playerx-1].item == ITEM_BOX) {
                    if (playerx > 1 &&
                        map[playery][playerx-2].background != BG_ROCK &&
                        map[playery][playerx-2].item != ITEM_BOX)
                    {
                        Backup1Level();
                        backup_done = 1;
                        map[playery][playerx-1].item  = ITEM_NONE;
                        map[playery][playerx-2].item  = ITEM_BOX;
                        map[playery][playerx-2].dirty = 1;
                    }
                    else {
                        continue;
                    }
                }
                if (!backup_done) Backup1Level();
                map[playery][playerx].dirty = 1;
                map[playery][playerx].item  = ITEM_NONE;
                playerx--;
                if (playerx-startx <= 1) {
                    if (startx > 0) {
                        startx--;
                        screen_refresh = 1;
                    }
                }
                map[playery][playerx].dirty = 1;
                map[playery][playerx].item  = ITEM_PLAYER;
            }
        }

        else if (input == KEY_RIGHT) {
            if (scrollmode) {
                if (startx < NR_COLUMNS - VISIBLE_COLUMNS) {
                    startx++;
                    screen_refresh = 1;
                }
                continue;
            }
            if (playerx < NR_COLUMNS - 1  && map[playery][playerx+1].background != BG_ROCK) {
                if (map[playery][playerx+1].item == ITEM_BOX) {
                    if (playerx < NR_COLUMNS-2 &&
                        map[playery][playerx+2].background != BG_ROCK &&
                        map[playery][playerx+2].item != ITEM_BOX)
                    {
                        Backup1Level();
                        backup_done = 1;
                        map[playery][playerx+1].item  = ITEM_NONE;
                        map[playery][playerx+2].item  = ITEM_BOX;
                        map[playery][playerx+2].dirty = 1;
                    }
                    else {
                        continue;
                    }
                }
                if (!backup_done) Backup1Level();
                map[playery][playerx].dirty = 1;
                map[playery][playerx].item  = ITEM_NONE;
                playerx++;
                if (playerx-startx >= VISIBLE_COLUMNS-2) {
                    if (startx < NR_COLUMNS - VISIBLE_COLUMNS) {
                        startx++;
                        screen_refresh = 1;
                    }
                }
                map[playery][playerx].dirty = 1;
                map[playery][playerx].item  = ITEM_PLAYER;
            }
        }

        else if (input == KEY_SAVE) {
            BackupLevel();
        }

        else if (input == KEY_BACK) {
            memcpy(map[0],back1_map[0],sizeof(sq_type)*NR_COLUMNS*NR_LINES);
            startx         = back1_startx;
            starty         = back1_starty;
            startx_old     = back1_startx_old;
            starty_old     = back1_starty_old;
            playerx        = back1_playerx;
            playery        = back1_playery;
            scrollmode     = back1_scrollmode;
            screen_refresh = 1;
        }

        else if (input == KEY_RESTORE) {
            memcpy(map[0],back_map[0],sizeof(sq_type)*NR_COLUMNS*NR_LINES);
            startx         = back_startx;
            starty         = back_starty;
            startx_old     = back_startx_old;
            starty_old     = back_starty_old;
            playerx        = back_playerx;
            playery        = back_playery;
            scrollmode     = back_scrollmode;
            screen_refresh = 1;
        }


        if ((finish = LevelCompleted())) RedrawDirtySquares();
    }
    while (input != KEY_ESCAPE && !finish);

    if (finish) return RequestDialog("Good. Try next level?");

    if (play_mode != RANDOM_MODE) {
        if (RequestDialog("Restart level?")) return 2;
        else                                 return 0;
    }

    return RequestDialog("Try another one?");
}



/*===========================================================================*/
/* main routine: where all the fun starts ...                                */
/*===========================================================================*/
void _main(void) {
    short ret_val;

    //--------------------------------------
    // try to allocate memory
    //--------------------------------------
    if (!InitMemory()) {
        FreeMemory();
        ClearScreen();
        DrawStr(0,0,"ERROR: not enough mem",A_REPLACE);
        GetInputKeyWithAPD();
        return;
    }

    //--------------------------------------
    // clear screen and output splash screen
    //--------------------------------------
    FXClearScreen();

    //--------------------------------------
    // turn on grayscale graphics subsystem
    //--------------------------------------
    if (!GrayOn()) {
        FreeMemory();
        return;
    }

    //--------------------------------------
    // fetch start addresses of planes
    //--------------------------------------
    plane1 = GetPlane(0);
    plane2 = GetPlane(1);

    //----------------------------------------
    // draws and animates title screen
    //----------------------------------------
    Logo();

    //--------------------------------------
    // initialize random number generator
    //--------------------------------------
    randomize();

    //--------------------------------------
    // loop until user aborts
    //--------------------------------------
    while (HandleMainMenu()) {
        do {
            SetPlane(1);
            ClearScreen();
            SetPlane(0);
            ClearScreen();
            ret_val = GenerateAndPlay();
            if (ret_val == 1 && play_mode != RANDOM_MODE) level_nr++;

        }
        while (ret_val);
    }

    //--------------------------------------
    // cleanup and bye =8-)
    //--------------------------------------
    FreeMemory();
    GrayOff();
}

